from uuid import uuid4, UUID
from datetime import datetime

from sqlmodel import SQLModel, Field

from sqlalchemy import func, text

# Tables for tracking a user's order.


class OrderTbl(SQLModel, table=True):
    __tablename__ = "order"

    id: UUID = Field(
        primary_key=True,
        index=True,
        default_factory=uuid4,
        sa_column_kwargs={"server_default": text("gen_random_uuid()")},
    )
    user_account_id: UUID
    # No foreign key since products could potentially be deleted but we'd want to maintain the order record.
    product_id: UUID
    created_at: datetime = Field(sa_column_kwargs={"server_default": func.now()})
    updated_at: datetime = Field(
        sa_column_kwargs={"server_default": func.now(), "onupdate": func.now()}
    )
    # PENDING (Awaiting payment), RECEIVED, COMPLETE, CANCELLED, REFUNDED
    order_status: str = Field(max_length=30, nullable=False)


# TODO implment into system
# class OrderExpenseTbl(SQLModel, table=True):
#     # Track expenses created while fulfilling an order.
#     __tablename__ = "order_expense"

#     id: UUID = Field(
#         UUID(as_uuid=True),
#         primary_key=True,
#         default=uuid4,
#         server_default=text("gen_random_uuid()"),
#         unique=True,
#         nullable=False,
#         index=True,
#     )
#     order_id: UUID = Field(
#         UUID(as_uuid=True),
#         ForeignKey("order.id"),
#     )
#     # Category e.g. 'rendering', 'file_storage', 'payment_processing', 'affiliate_payment'
#     expense_type: str = Field(String(30), nullable=False)
#     # Specific info e.g. 'stripe fees' or '300 credits via Conductor' or 'Project files (12.2gb) stored since 01/01/2024'
#     expense_description: str = Field(String(200), nullable=False)


# TODO implement into system
# class OrderRevenueTbl(SQLModel, table=True):
#     # Track sources of revenue related to this order,
#     # or if the order was payed for with credits/tokens?
#     __tablename__ = "order_payment"

#     id: UUID = Field(
#         UUID(as_uuid=True),
#         primary_key=True,
#         default=uuid4,
#         server_default=text("gen_random_uuid()"),
#         unique=True,
#         nullable=False,
#         index=True,
#     )
#     order_id: UUID = Field(
#         UUID(as_uuid=True),
#         ForeignKey("order.id"),
#         primary_key=True,
#     )
#     # Stripe / other
#     payment_provider: str = Field(String(30))



import logging
from sqlmodel import Session
from ceonmedia_db import models

from pydantic import BaseModel

logger = logging.getLogger(__name__)


class ProductCreate(BaseModel):
    name: str
    description: str
    price_gbp_pence: int


DEFAULT_PRODUCTS = [
    ProductCreate(
        name="video_preview",
        description="Create a watermarked preview video using your custom inputs. Non commercial use.",
        price_gbp_pence=5_00,
    ),
    ProductCreate(
        name="video_hd",
        description="Create a video in HD 1920 x 1080 resolution, using your custom inputs. Includes a Ceon Media Commercial License.",
        price_gbp_pence=250_00,
    ),
    ProductCreate(
        name="video_4k",
        description="Create a video in 4K 3840 x 2160 resolution, using your custom inputs. Includes a Ceon Media Commercial License",
        price_gbp_pence=400_00,
    ),
]


def create_default_products(session: Session):
    logger.info("Adding default products to database...")
    product_sqlalchemy_instances = [
        models.product.ProductTbl(**default_product.__dict__)
        for default_product in DEFAULT_PRODUCTS
    ]
    session.add_all(product_sqlalchemy_instances)
    logger.info("Added default products to session.")
    session.commit()
    logger.info("Committed default products to database.")



from uuid import UUID, uuid4
from datetime import datetime

from sqlmodel import SQLModel, Field, Relationship

from sqlalchemy import (
    func,
    text,
)

from . import access_control


class UserTbl(SQLModel, table=True):
    __tablename__ = "user"

    id: UUID = Field(
        primary_key=True,
        default_factory=uuid4,
    )
    username: str | None
    email: str = Field(max_length=255, index=True)
    image: str | None
    created_at: datetime = Field(
        default_factory=datetime.utcnow,
        sa_column_kwargs={"onupdate": func.now(), "server_default": func.now()},
    )
    updated_at: datetime = Field(
        default_factory=datetime.utcnow,
        sa_column_kwargs={"onupdate": func.now(), "server_default": func.now()},
    )

    # Relationships
    user_projects: list["ProjectTbl"] = Relationship(back_populates="owner")
    user_roles: list["RoleTbl"] = Relationship(  # type: ignore
        link_model=access_control.UserRoleLinkTbl
    )
    accounts: list["UserAccountTbl"] = Relationship(back_populates="user")
    reusable_preview_credits: list["ReusablePreviewTokenTbl"] = Relationship(
        back_populates="user"
    )

    def __repr__(self):
        return f"User(id={self.id!r}, email={self.email!r})"


# ------------------
# -- Child tables --
# ------------------


# One user may have multiple 'accounts' (login methods e.g. Google, Github, Email)
class UserAccountTbl(SQLModel, table=True):
    __tablename__ = "oauth_account"

    # id: Mapped[int] = mapped_column(primary_key=True)
    id: UUID = Field(
        primary_key=True,
        index=True,
        default_factory=uuid4,
        sa_column_kwargs={
            "server_default": text("gen_random_uuid()"),
        },
    )
    user_id: UUID = Field(
        foreign_key="user.id",
        ondelete="CASCADE",
    )
    provider_id: str = Field(max_length=255)
    provider_user_id: str = Field(max_length=255)

    user: "UserTbl" = Relationship(back_populates="accounts")



# Test
import logging
import time
from uuid import UUID

from sqlmodel import Session
from sqlalchemy import exc
from sqlalchemy import select

from ceonmedia_db.models import user as user_models

# from ceonmedia_db.auth import security
from ceonmedia_db import errors
from .oauth_account import OAuthProvider, OAuthProviderValue
from ceonmedia_db import constants

logger = logging.getLogger(constants.LOGGER_NAME)

# Currently not being used since api almost always starts interaction with an email get
# Leaving here in case it needs to be reactivated
# def get(db: Session, user_uuid: UUID) -> user_models.UserTbl:
#     logger.debug(f"Getting user by id: {user_uuid}")
#     stmt = select(user_models.UserTbl).where(
#         user_models.UserTbl.id == user_uuid
#     )
#     try:
#         got_user = db.execute(stmt).scalars().one()
#     except exc.NoResultFound:
#         raise errors.NotFoundInDBError(f"Could not find user by uuid: {uuid}")
#     logger.debug(f"Returning: {got_user}")
#     if not got_user:
#         raise errors.NotFoundInDBError(f"User {user_uuid} not found in DB")
#     return got_user


def get(db: Session, user_uuid: UUID | str) -> user_models.UserTbl:
    start_time = time.time()
    stmt = select(user_models.UserTbl).where(user_models.UserTbl.id == user_uuid)
    try:
        got_user = db.execute(stmt).scalars().one()
    except exc.NoResultFound:
        raise errors.NotFoundInDBError(f"Could not find user by uuid: {user_uuid}")
    logger.warning(f"DB user fetch: {time.time() - start_time:.3f} seconds")
    logger.debug(f"Returning: {got_user}")
    return got_user


def get_by_email(db: Session, email: str) -> user_models.UserTbl:
    stmt = select(user_models.UserTbl).where(user_models.UserTbl.email == email)
    try:
        got_user = db.execute(stmt).scalars().one()
    except exc.NoResultFound:
        raise errors.NotFoundInDBError(f"Could not find user by email: {email}")
    logger.debug(f"Returning: {got_user}")
    return got_user


def get_all(db: Session):
    stmt = select(user_models.UserTbl).order_by(user_models.UserTbl.email)
    result = db.execute(stmt).scalars().all()
    logger.debug(f"result: {result}")
    return result


def create(
    db: Session,
    email: str,
    image: str | None = None,
    username: str | None = None,
) -> user_models.UserTbl:
    """Create a user"""
    logger.debug(f"Creating new user: {email}")
    new_user_account = user_models.UserTbl(email=email, image=image, username=username)
    logger.debug(f"Adding new user: {new_user_account}")
    db.add(new_user_account)
    logger.debug("Flushing new user.")
    db.flush()
    return new_user_account


def create_oauth_account(
    db: Session,
    user_uuid: str | UUID,
    provider_id: OAuthProviderValue | OAuthProvider,
    provider_user_id: str,
) -> user_models.UserAccountTbl:
    """Create an oauth account for the target user."""
    # Enforce valid enum by loading as enum instance
    provider_id = OAuthProvider(provider_id).value
    logger.debug(
        f"Creating new oauth account: {user_uuid=} {provider_id=}, {provider_user_id=}"
    )
    new_oauth_account = user_models.UserAccountTbl(
        user_id=user_uuid,
        provider_id=provider_id,
        provider_user_id=provider_user_id,
    )
    logger.debug(f"Adding new oauth_account: {new_oauth_account}")
    db.add(new_oauth_account)
    return new_oauth_account


# Moved sessions handling to REDIS
# def create_session(db, user_uuid: str | UUID):
#     new_session = user_models.UserSessionsTbl(user_id=user_uuid)
#     logger.debug(f"db.add() new user session: {new_session}")
#     db.add(new_session)
#     return new_session



# TODO zone out (slowly remove) schamas here. Leave this responsability to the API.
from .job import JobCreate
from .project import ProjectCreate
from .user import UserAccountCreate, UserProfileCreate
from .access_control import PermissionOption, UserRoleOption



from enum import Enum


class UserRoleOption(str, Enum):
    VERIFIED_USER = "verified"
    ARTIST = "artist"
    AFFILIATE = "affiliate"
    ADMIN = "admin"


class PermissionOption(str, Enum):
    SUBMIT_PREVIEWS = "submit_preview"
    PURCHASE = "purchase"
    MANAGE_PROJECT = "manage_project"
    DELETE_PROJECT = "delete_project"



import logging
from typing import Union
from uuid import UUID

from sqlalchemy import select
from sqlalchemy import exc
from sqlmodel import Session

from ceonmedia_db import models
from ceonmedia_db import errors
from ceonmedia_db import constants


logger = logging.getLogger(constants.LOGGER_NAME)


def create_email_verification_token(
    db: Session, user_account_uuid: Union[UUID, str], email_to_verify: str
):
    logger.debug("Creating temporary security token...")
    new_token = models.security.EmailVerificationTokenTbl(
        user_account_uuid=user_account_uuid, email_to_verify=email_to_verify
    )
    db.add(new_token)
    logger.debug(f"Added new token to session: {new_token}")

    # Commit the changes and return the new token
    try:
        db.commit()
    except exc.SQLAlchemyError as e:
        msg = f"Failed to commit changes (creating token) for user: {user_account_uuid}"
        logger.error(msg)
        errors.handle_sql_error(e, msg=msg)
    except Exception as e:
        logger.error("Unknown exception")
        raise e
    return new_token


def get_email_verification_token(db: Session, token_uuid: Union[UUID, str]):
    # Get the token from the database
    stmt = select(models.security.EmailVerificationTokenTbl).where(
        models.security.EmailVerificationTokenTbl.id == token_uuid
    )
    token = db.execute(stmt).scalars().first()
    if not token:
        raise errors.NotFoundInDBError(
            f"Email verification token not found in db: {token_uuid}"
        )
    return token


def redeem_email_verification_token(db: Session, token_uuid: Union[UUID, str]):
    # TODO check that token is not expired
    """
    Executes 'email is verified' actions.
    - Update the user's main email address
    - Mark the user email as confirmed <--------
    - TODO assing verified user role <- TODO don't need both of these? Better to jsut use roles?
    - Delete the email-verifcation token.
    """
    # Get the token from the database
    stmt = select(models.security.EmailVerificationTokenTbl).where(
        models.security.EmailVerificationTokenTbl.id == token_uuid
    )
    token = db.execute(stmt).scalars().first()
    if not token:
        raise errors.NotFoundInDBError(
            f"Email verification token not found in db: {token_uuid}"
        )

    logger.debug("Getting user account tbl...")
    stmt = select(models.user.UserAccountTbl).where(
        models.user.UserAccountTbl.id == token.user_account_uuid
    )
    user_account = db.execute(stmt).scalars().first()
    if not user_account:
        raise errors.NotFoundInDBError(
            f"User not found in db for email verification token {token_uuid}: User({token.user_account_uuid})"
        )

    # Update the user fields
    user_account.email = token.email_to_verify
    user_account.email_is_confirmed = True

    logger.warning("Deleting used token...")
    db.delete(token)

    # -- Commit Changes --
    try:
        db.commit()
    except exc.SQLAlchemyError as e:
        msg = f"Failed to commit changes (verifying user email) for user: {user_account.email}"
        errors.handle_sql_error(e, msg=msg)
    logger.info(
        f"Committed changes to db (verified email) for user: {user_account.email}"
    )



# from .job_models import JobTbl
# from .job_models import JobProgressTbl

from .user import UserAccountTbl

# from .user import UserProfileTbl
# from .user import UserLoginTbl
# from .security import EmailVerificationTokenTbl

from .access_control import RoleTbl
from .access_control import PermissionTbl

from .access_control import RolePermissionLinkTbl
from .access_control import UserRoleLinkTbl

# Payments and credits
from .user_credit import ReusablePreviewTokenTbl
from .stripe import UserStripeTbl

from .project import ProjectTbl
from .submission import SubmissionTbl
from .job import JobTbl
from .job import JobProgressTbl

from .order import OrderTbl
from .product import ProductTbl



import pytest
import logging

from ceonmedia_db import crud
from ceonmedia_db import schemas
from ceonmedia_db import errors

logger = logging.getLogger(__name__)


def test_create_user_role_success(empty_db_session):
    with empty_db_session() as session:
        existing_roles = crud.access_control.get_all_user_roles(session)
        logger.warning(f"Found existing user roles: {existing_roles}")
        role_to_create = schemas.UserRoleOption.VERIFIED_USER
        crud.access_control.create_role(session, role_to_create)
        session.commit()


def test_create_user_role_already_exists(empty_db_session):
    # First create the original entry
    with empty_db_session() as session:
        role_to_create = schemas.UserRoleOption.VERIFIED_USER
        crud.access_control.create_role(session, role_to_create)
        with pytest.raises(errors.AlreadyExistsInDBError):
            crud.access_control.create_role(session, role_to_create)


def test_add_permission_to_role_success(default_db_session):
    user_role = schemas.UserRoleOption.AFFILIATE
    permission = schemas.PermissionOption.SUBMIT_PREVIEWS
    with default_db_session() as session:
        crud.access_control.add_permission_to_role(session, user_role, permission)


def test_add_permission_to_role_already_exists(default_db_session):
    user_role = schemas.UserRoleOption.AFFILIATE
    permission = schemas.PermissionOption.SUBMIT_PREVIEWS
    with default_db_session() as session:
        crud.access_control.add_permission_to_role(session, user_role, permission)
        with pytest.raises(errors.AlreadyExistsInDBError):
            crud.access_control.add_permission_to_role(session, user_role, permission)


def test_get_permissions_for_user_role(default_db_session):
    user_role = schemas.UserRoleOption.VERIFIED_USER
    with default_db_session() as session:
        for user_role in schemas.UserRoleOption:
            list_of_permission_enums = (
                crud.access_control.get_all_permissions_for_user_role(
                    session, user_role
                )
            )
            logger.warning(
                f"Got permissions for role {user_role}:\n{list_of_permission_enums}"
            )
            for permission_enum in list_of_permission_enums:
                # Ensure returned type is an enum instance
                assert permission_enum in schemas.PermissionOption



import logging

# from sqlalchemy import create_engine
# from sqlmodel import Sessionmaker

logger = logging.getLogger(__name__)

DEFAULT_CONNECT_ARGS = {"sslmode": "require"}
connect_args_local_docker = {}


# For fastapi to create db sessions for each request
# def create_local_session_maker(db_uri: str, connect_args=None) -> sessionmaker:
#     if connect_args == None:
#         connect_args = DEFAULT_CONNECT_ARGS
#     logger.warning(
#         "TODO make sure to use connect_args with sslmode require in production"
#     )
#     logger.warning(f"{connect_args=}")
#     engine = create_engine(
#         # DB_URI, connect_args=connect_args, pool_pre_ping=True
#         db_uri,
#         connect_args=connect_args,
#         pool_pre_ping=True,
#     )
#     SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)
#     return SessionLocal


def test_db_engine_connection(engine):
    logger.info("Testing DB connection...")
    logger.info(f"URI: {engine.url}")
    conn = engine.connect()
    conn.close()
    logger.info(f"Closed test connection: {engine.url}")


def test_db_session_connection(session):
    engine = session.get_bind()
    logger.info("Testing DB connection...")
    logger.info(f"URI: {engine.url}")
    conn = engine.connect()
    conn.close()
    logger.info(f"Closed test connection: {engine.url}")


def get_db_uri_from_session(session):
    engine = session.get_bind()
    return engine.url



from fastapi import Request, HTTPException
from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials

from .auth_handler import decodeJWT


class JWTBearer(HTTPBearer):
    def __init__(self, auto_error: bool = True):
        super(JWTBearer, self).__init__(auto_error=auto_error)

    async def __call__(self, request: Request):
        credentials: HTTPAuthorizationCredentials = await super(
            JWTBearer, self
        ).__call__(request)
        if credentials:
            if not credentials.scheme == "Bearer":
                raise HTTPException(
                    status_code=403, detail="Invalid authentication scheme."
                )
            if not self.verify_jwt(credentials.credentials):
                raise HTTPException(
                    status_code=403, detail="Invalid token or expired token."
                )
            return credentials.credentials
        else:
            raise HTTPException(status_code=403, detail="Invalid authorization code.")

    def verify_jwt(self, jwtoken: str) -> bool:
        isTokenValid: bool = False

        try:
            payload = decodeJWT(jwtoken)
        except:
            payload = None
        if payload:
            isTokenValid = True
        return isTokenValid



import os
import logging
from dotenv import load_dotenv
from ceonmedia_db import database

load_dotenv()
logger = logging.getLogger(__name__)

TEST_DB_URI = os.environ["TEST_DB_URI"]
connect_args = {}  # No ssl for local dev connecton

TestSessionLocal = database.create_local_session_maker(
    TEST_DB_URI, connect_args=connect_args
)



# Authentication
import logging

from ceonmedia_db.crud import user as user_crud

# Moved inside functions to avoid dependencies when importing ceonmedia_db
# from passlib.context import CryptContext
# from jose import JWTError, jwt

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# ALGORITHM = "HS256"
# ACCESS_TOKEN_EXPIRE_MINUTES = 60 * 24


def authenticate_credentials(db, email: str, unhashed_password: str):
    # Importing here to avoid dependencies when ceonmedia_db is imported by render/submisssion servers
    from passlib.context import CryptContext

    pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")

    # verify password
    saved_hash = user_crud.Login.get_saved_hash(db, email)
    logger.debug(f"got login for email: {email}:{saved_hash}")
    if not saved_hash:
        raise Exception(
            "Failed to find hashed_password in DB when trying to authenticate user credentials"
        )
    if pwd_context.verify(unhashed_password, saved_hash):
        user = user_crud.User.get_by_email(db, email)
        logger.debug(f"user: {user}")
        return user
    return None


def hash_password(password):
    from passlib.context import CryptContext

    pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
    return pwd_context.hash(password)


""" Not DB related. Move to API server code
def create_access_token(payload: dict):
    logger.debug(f"received payload: {payload}")
    restructured_data = {"sub": payload}
    logger.debug(f"restructured data: {restructured_data}")
    to_encode = restructured_data.copy()
    expire = datetime.utcnow() + timedelta(minutes=ACCESS_TOKEN_EXPIRE_MINUTES)
    to_encode.update({"exp": expire})
    logger.debug(f"added 'exp' to data: {to_encode}")
    if not JWT_SECRET:
        msg = "Unable to create token due to missing ENV var: JWT_SECRET"
        logger.error(msg)
        raise Exception(msg)
    encoded_jwt = jwt.encode(to_encode, JWT_SECRET, algorithm=ALGORITHM)
    logger.debug(f"encoded_jwt: {encoded_jwt}")
    return encoded_jwt
"""


# WIP remove api/website functionality from this module.
# def create_access_token_cookie(payload):
#     max_age_seconds = 60 * ACCESS_TOKEN_EXPIRE_MINUTES
#     access_token = create_access_token(payload)
#     domain = "localhost" if (os.getenv("ENV") == "DEV") else "ceonmedia.com"
#     access_token_cookie = f"access_token={access_token}; Domain={domain}; Path=/; Max-Age={max_age_seconds}; HttpOnly"
#     logger.debug(
#         f"Created access token cookie with payload({payload}): {access_token_cookie}"
#     )
#     return access_token_cookie


# WIP remove api/website functionality from this module.
# def decode_access_token(token):
#     payload = jwt.decode(token, JWT_SECRET, algorithms=[ALGORITHM])
#     logger.debug(f"Decoded token: {token}\nToken payload: {payload}")
#     return payload



# Test
import logging
from enum import Enum
from typing_extensions import Literal
from typing import get_args

from sqlmodel import Session
from sqlalchemy import exc
from sqlalchemy import select

from ceonmedia_db.models import user as user_models

# from ceonmedia_db.auth import security
from ceonmedia_db import errors
from ceonmedia_db import constants

logger = logging.getLogger(constants.LOGGER_NAME)


class OAuthProvider(str, Enum):
    GOOGLE = "google"
    GITHUB = "github"
    EMAIL = "email"


# Created to type hint for allowing string values for static type checking
OAuthProviderValue = Literal["google", "github", "email"]

# Enforce that type hinting is synch'd with enum
assert set(get_args(OAuthProviderValue)) == {member.value for member in OAuthProvider}


def get(
    db: Session,
    provider_id: OAuthProviderValue | OAuthProvider,
    provider_user_id: str,
) -> user_models.UserAccountTbl:
    if isinstance(provider_id, OAuthProvider):
        provider_id = provider_id.value
    stmt = select(user_models.UserAccountTbl).where(
        user_models.UserAccountTbl.provider_id == str(provider_id),
        user_models.UserAccountTbl.provider_user_id == str(provider_user_id),
    )
    try:
        got_user_account = db.execute(stmt).scalars().one()
    except exc.NoResultFound:
        raise errors.NotFoundInDBError(
            f"Could not find user_account: {provider_id=}, {provider_user_id=}"
        )
    logger.debug(f"Returning: {got_user_account}")
    return got_user_account



import logging

from sqlalchemy import select, update
from sqlmodel import Session
from sqlalchemy.exc import NoResultFound

# from sqlalchemy.sql import values
from uuid import UUID

from ceonmedia_db.models.project_models import ProjectTbl

# from db.engine import SessionLocal

logger = logging.getLogger(__name__)


class Project:
    @classmethod
    def get(cls, db: Session, proj_uuid: UUID, raise_missing=False):
        stmt = select(ProjectTbl).where(ProjectTbl.uuid == proj_uuid)
        try:
            got_project = db.execute(stmt).scalar_one()
        except NoResultFound as e:
            if raise_missing:
                raise e
            return None
        return got_project

    @classmethod
    def get_all(cls, db: Session):
        # TODO only return jobs belonging to user id
        stmt = select(ProjectTbl).order_by(ProjectTbl.title)
        result = db.execute(stmt).scalars().all()
        return result

    @classmethod
    def get_proj_inputs(cls, db: Session, proj_uuid: UUID):
        stmt = select(ProjectTbl.proj_inputs).where(ProjectTbl.uuid == proj_uuid)
        got_project_inputs = db.execute(stmt).one()
        # return got_project["proj_inputs"]
        return got_project_inputs.proj_inputs

    @classmethod
    def create(cls, db: Session, **values):
        """Create a new project row"""
        new_project = ProjectTbl(**values)
        db.add(new_project)
        db.commit()
        return new_project

    @classmethod
    def update(cls, db: Session, proj_uuid: UUID, **values):
        """Update the object or return None if not found."""
        project = db.get(ProjectTbl, proj_uuid)
        if not project:
            logger.info("--NO PROJECT FOUND--")
            return None
        logger.info(f"Got from DB project.uuid({project.uuid}): {project.title}")
        logger.info(f"Got values: {values}")
        # project.update(**values)
        stmt = update(ProjectTbl).where(ProjectTbl.uuid == proj_uuid).values(**values)
        db.execute(stmt)
        # rowcount = result.rowcount
        # logger.info(f"update project result: {result}")
        # logger.info(f"update project rowcount: {rowcount}")
        db.commit()
        return project

    @classmethod
    def upsert(cls, db: Session, proj_uuid: UUID, **values):
        """Update the object or create a new entry if not found."""
        new_project = ProjectTbl(uuid=proj_uuid, **values)
        db.merge(new_project)
        db.commit()

    @classmethod
    def delete(cls, db: Session, proj_uuid: UUID):
        project = db.get(ProjectTbl, proj_uuid)
        logger.info(f"Got project for deletion: {project}")
        db.delete(project)
        db.commit()
        logger.info(f"Deleted proj({proj_uuid}) from db.")
        return proj_uuid



import logging

from sqlalchemy import select
from sqlmodel import Session
from sqlalchemy.exc import NoResultFound, IntegrityError

# from sqlalchemy.sql import values
from uuid import UUID

from ceonmedia_db import models
from ceonmedia_db import schemas
from ceonmedia_db import errors
from ceonmedia_db import constants

# from db.engine import SessionLocal


logger = logging.getLogger(constants.LOGGER_NAME)


def delete(db: Session, project_uuid: UUID):
    project_to_delete = (
        db.query(models.ProjectTbl).filter(models.ProjectTbl.id == project_uuid).first()
    )
    db.delete(project_to_delete)
    db.commit()
    logger.info("Deleted project: %s", project_uuid)


def create(db: Session, project_create: schemas.ProjectCreate):
    logger.info(f"Creating project with project_create:\n{project_create}")
    new_project = models.ProjectTbl(**project_create.dict())
    logger.info(f"Adding project model to db:\n{new_project}")
    db.add(new_project)
    logger.info("Commiting changes to db...")
    db.commit()
    return new_project


def get(session: Session, project_uuid: UUID) -> models.ProjectTbl:
    stmt = select(models.ProjectTbl).where(models.ProjectTbl.id == project_uuid)
    try:
        got_project = session.execute(stmt).scalar_one()
    except NoResultFound as e:
        raise errors.NotFoundInDBError(f"Project {project_uuid} not found in db: {e}")
    return got_project


def get_proj_inputs(session: Session, project_uuid: UUID):
    stmt = select(models.ProjectTbl).where(models.ProjectTbl.id == project_uuid)
    try:
        got_project = session.execute(stmt).scalar_one()
    except NoResultFound as e:
        raise errors.NotFoundInDBError(f"Project {project_uuid} not found in db: {e}")
    proj_inputs = got_project.proj_inputs
    return proj_inputs


def get_all(db: Session):
    # TODO implement skip/pagination
    # TODO only return jobs belonging to user id
    logger.warning("TODO implement skipping/pagination for get_all projects")
    stmt = select(models.ProjectTbl).order_by(models.ProjectTbl.title)
    result = db.execute(stmt).scalars().all()
    return result


# def update(db: Session, proj_uuid: UUID, **values):
#     project = db.get(models.ProjectTbl, proj_uuid)
#     if not project:
#         logger.info("--NO PROJECT FOUND--")
#         return None
#     logger.info(f"Got from DB project.id({project.id}): {project.title}")
#     logger.info(f"Got values: {values}")
#     # project.update(**values)
#     stmt = update(models.ProjectTbl).where(models.ProjectTbl.id == proj_uuid).values(**values)
#     db.execute(stmt)
#     # rowcount = result.rowcount
#     # logger.info(f"update project result: {result}")
#     # logger.info(f"update project rowcount: {rowcount}")
#     db.commit()
#     return project


def upsert(db: Session, proj_uuid: UUID, **values):
    new_project = models.ProjectTbl(uuid=proj_uuid, **values)
    try:
        db.merge(new_project)
        db.commit()
    except IntegrityError as e:
        errors.handle_sql_error(e)


"""
class Project:
    @classmethod
    def get(cls, db: Session, proj_uuid: UUID, raise_missing=False):
        stmt = select(ProjectTbl).where(ProjectTbl.id == proj_uuid)
        try:
            got_project = db.execute(stmt).scalar_one()
        except NoResultFound as e:
            if raise_missing:
                raise e
            return None
        return got_project

    @classmethod
    def get_all(cls, db: Session):
        # TODO only return jobs belonging to user id
        stmt = select(ProjectTbl).order_by(ProjectTbl.title)
        result = db.execute(stmt).scalars().all()
        return result

    @classmethod
    def get_proj_inputs(cls, db: Session, proj_uuid: UUID):
        stmt = select(ProjectTbl.proj_inputs).where(ProjectTbl.id == proj_uuid)
        got_project_inputs = db.execute(stmt).one()
        # return got_project["proj_inputs"]
        return got_project_inputs.proj_inputs

    @classmethod
    def create(cls, db: Session, **values):
        new_project = ProjectTbl(**values)
        db.add(new_project)
        db.commit()
        return new_project

    @classmethod
    def update(cls, db: Session, proj_uuid: UUID, **values):
        project = db.get(ProjectTbl, proj_uuid)
        if not project:
            logger.info("--NO PROJECT FOUND--")
            return None
        logger.info(f"Got from DB project.id({project.id}): {project.title}")
        logger.info(f"Got values: {values}")
        # project.update(**values)
        stmt = update(ProjectTbl).where(ProjectTbl.id == proj_uuid).values(**values)
        db.execute(stmt)
        # rowcount = result.rowcount
        # logger.info(f"update project result: {result}")
        # logger.info(f"update project rowcount: {rowcount}")
        db.commit()
        return project

    @classmethod
    def upsert(cls, db: Session, proj_uuid: UUID, **values):
        new_project = ProjectTbl(uuid=proj_uuid, **values)
        db.merge(new_project)
        db.commit()

    @classmethod
    def delete(cls, db: Session, proj_uuid: UUID):
        project = db.get(ProjectTbl, proj_uuid)
        logger.info(f"Got project for deletion: {project}")
        db.delete(project)
        db.commit()
        logger.info(f"Deleted proj({proj_uuid}) from db.")
        return proj_uuid
"""



import logging
from typing import Sequence

from sqlalchemy import select, update, desc
from sqlmodel import Session

# from sqlalchemy.sql import values
from uuid import UUID

from ceonmedia_db import models
from ceonmedia_db import constants

# from ceonmedia_db.models.job_models import (
#     JobTbl,
#     JobProgressTbl,
# )

# from db.engine import SessionLocal


logger = logging.getLogger(constants.LOGGER_NAME)


# def create(db: Session, **values):
def create(
    db: Session,
    user_id: UUID,
    project_id: UUID,
    submission_id: UUID,
    job_type: str,
):
    # Create new job entry
    new_job = models.JobTbl(
        user_account_id=user_id,
        project_id=project_id,
        submission_id=submission_id,
        job_type=job_type,
    )
    db.add(new_job)
    db.flush()
    logger.debug(f"Created new_job: {new_job}")
    # Create new job_progress entry
    new_job_progress = models.JobProgressTbl(job_id=new_job.id)
    db.add(new_job_progress)
    logger.debug(f"Created new_job_progress: {new_job_progress}")
    db.commit()

    return new_job


def get(db: Session, job_id) -> models.JobTbl:
    stmt = select(models.JobTbl).where(models.JobTbl.id == job_id)
    got_job = db.execute(stmt).scalars().one()
    return got_job


def get_unstarted_jobs(
    db: Session,
) -> Sequence[models.JobProgressTbl]:
    stmt = select(models.JobProgressTbl).where(
        models.JobProgressTbl.assigned_at == None
    )
    got_job = db.execute(stmt).scalars().all()
    return got_job


def get_job_progress(db: Session, job_id):
    stmt = select(models.JobProgressTbl).where(models.JobProgressTbl.job_id == job_id)
    got_job_progress = db.execute(stmt).scalars().one()
    return got_job_progress


def update_job(db: Session, job_id, values_to_set: dict):
    if not values_to_set:
        logger.warning("Received blank values_to_set. Skipping")
        return None
    stmt = (
        update(models.JobTbl).where(models.JobTbl.id == job_id).values(**values_to_set)
    )
    try:
        # Note: Changes not committed to db until session.flush()/commit() is executed.
        # This logic is handled by the app/API so that DB calls can be minimized
        db.execute(stmt)
        return
    except Exception as e:
        logger.warn(f"Failed to set timestamp in jobqueue: {e}")
        raise Exception(f"Failed to update jobqueue timestamps: {values_to_set}")


def update_job_progress(db: Session, job_id, values_to_set: dict):
    # job_id: The id of the job to update
    # columns_to_timestamp: A list of strings that are the column names
    if not values_to_set:
        logger.warning("Received blank values_to_set. Skipping")
        return None

    # timestamp = datetime.now()
    msgs_to_print = ["Got values_to_set: "]
    for key, value in values_to_set.items():
        msgs_to_print.append(f"\n\t\t{key}: {value}")
    logger.debug("".join(msgs_to_print))
    stmt = (
        update(models.JobProgressTbl)
        .where(models.JobProgressTbl.job_id == job_id)
        .values(**values_to_set)
    )
    try:
        # Note: Changes not committed to db until session.flush()/commit() is executed.
        # This logic is handled by the app/API so that DB calls can be minimized
        result = db.execute(stmt)
        logger.debug(
            f"Succesfully updated job progress (not committed): {values_to_set}"
        )
        logger.info(f"Got result from db: {result}")
        return values_to_set
    except Exception as e:
        logger.warn(f"Failed to set timestamp in jobqueue: {e}")
        raise Exception(f"Failed to update jobqueue timestamps: {values_to_set}")


def get_all(db: Session, page=0, limit=15):
    skip = page * limit
    stmt = (
        select(models.JobTbl)
        .offset(skip)
        .limit(limit)
        .order_by(desc(models.JobTbl.created_at))
    )
    result = db.execute(stmt).scalars().all()
    return result


def get_jobs_by_user_id(db: Session, user_id, page=0, limit=15):
    skip = page * limit
    stmt = (
        select(models.JobTbl)
        .where(models.JobTbl.user_account_id == user_id)
        .offset(skip)
        .limit(limit)
        .order_by(desc(models.JobTbl.created_at))
    )
    result = db.execute(stmt).scalars().all()
    return result


"""
class Job:
    @classmethod
    def get(cls, db: Session, job_id):
        stmt = select(models.JobTbl).where(models.JobTbl.id == job_id)
        got_user = db.execute(stmt).scalars().one()
        return got_user

    @classmethod
    def get_all(cls, db: Session, page=0, limit=15):
        skip = page * limit
        stmt = (
            select(models.JobTbl)
            .offset(skip)
            .limit(limit)
            .order_by(desc(models.JobTbl.id))
        )
        result = db.execute(stmt).scalars().all()
        return result

    @classmethod
    def get_jobs_by_user_id(cls, db: Session, user_id, page=0, limit=15):
        skip = page * limit
        stmt = (
            select(models.JobTbl)
            .where(models.JobTbl.user_id == user_id)
            .offset(skip)
            .limit(limit)
            .order_by(desc(models.JobTbl.created_at))
        )
        result = db.execute(stmt).scalars().all()
        return result

    # TODO refactor above func so filtering (see below func) can be an optional argument

    @classmethod
    def get_jobs_by_job_id_list(
        cls, db: Session, user_id, job_id_list, page=0, limit=15
    ):
        # returns a user's jobs, filtered by a list of job ids
        skip = page * limit
        stmt = (
            select(models.JobTbl)
            # .where(models.JobTbl.user_id == user_id)
            .where(
                and_(
                    models.JobTbl.user_id == user_id,
                    models.JobTbl.id.in_(job_id_list),
                )
            )
            .offset(skip)
            .limit(limit)
        )
        result = db.execute(stmt).scalars().all()
        return result

    @classmethod
    def create(cls, db: Session, **values):
        # Create new job entry
        new_job = models.JobTbl(**values)
        db.add(new_job)
        db.flush()
        logger.debug(f"Created new_job: {new_job}")
        # Create new job_progress entry
        new_job_progress = JobProgressTbl(job_id=new_job.id)
        db.add(new_job_progress)
        logger.debug(f"Created new_job_progress: {new_job_progress}")
        db.commit()

        return new_job


class JobProgress:
    @classmethod
    def get(cls, db: Session, job_id):
        stmt = select(JobProgressTbl).where(JobProgressTbl.job_id == job_id)
        got_job = db.execute(stmt).scalars().one()
        return got_job

    @classmethod
    def get_all(cls, db: Session):
        # TODO only return jobs belonging to user id
        stmt = select(JobProgressTbl).order_by(desc(JobProgressTbl.job_id))
        result = db.execute(stmt).scalars().all()
        return result

    @classmethod
    def create(cls, db: Session, **values):
        new_job_progress = JobProgressTbl(**values)
        db.add(new_job_progress)
        db.commit()
        return new_job_progress

    @classmethod
    def set_timestamp(cls, db: Session, job_id, columns_to_timestamp):
        # job_id: The id of the job to update
        # columns_to_timestamp: A list of strings that are the column names
        if not columns_to_timestamp:
            return None
        if not type(columns_to_timestamp) is list:
            columns_to_timestamp = [columns_to_timestamp]

        timestamp = datetime.now()
        msgs_to_print = ["Got keys to set: "]
        values_to_set = {}
        for key in columns_to_timestamp:
            msgs_to_print.append(f"\n\t\t{key}")
            values_to_set[key] = timestamp
        logger.info("".join(msgs_to_print))
        stmt = (
            JobProgressTbl.update()
            .where(JobProgressTbl.id == job_id)
            .values(**values_to_set)
        )
        try:
            result = db.execute(stmt).scalars().one()
            logger.info(f"Succesfully set timestamps: {values_to_set}")
            return values_to_set
        except Exception as e:
            logger.warn(f"Failed to set timestamp in jobqueue: {e}")
            raise Exception(f"Failed to update jobqueue timestamps: {values_to_set}")
"""



from uuid import uuid4, UUID

from sqlmodel import SQLModel, Field, Relationship

# import sqlalchemy
from sqlalchemy import (
    text,
)
# from ceonmedia_db.models.base import Base

# Permissions setup following the advice of:
# https://vertabelo.com/blog/user-authentication-module/
# With SQLAlchemy's recommended implementation for mant-to-many relationships, using association tables:
# https://docs.sqlalchemy.org/en/20/orm/basic_relationships.html#many-to-many

# --------------------
# -- Mapping Tables --
# --------------------
# To create many:many relationships between entities we use a third table for mapping


# note for a Core table, we use the sqlalchemy.Column construct,
# not sqlalchemy.orm.mapped_column
class RolePermissionLinkTbl(SQLModel, table=True):
    __tablename__ = "granted_permission"

    role_id: UUID = Field(default=None, primary_key=True, foreign_key="role.id")
    permission_id: UUID = Field(
        default=None, primary_key=True, foreign_key="permission.id"
    )


# granted_permission_table = Table(
#     # Assigns permissions to user_roles
#     "granted_permission",
#     Base.metadata,
#     Column("role_uuid", ForeignKey("role.uuid"), primary_key=True),
#     Column("permission_uuid", ForeignKey("permission.uuid"), primary_key=True),
#     UniqueConstraint(  # Don't allow entries with the same user_role/permission combination
#         "role_uuid", "permission_uuid", name="_granted_permission_uc"
#     ),
# )


class UserRoleLinkTbl(SQLModel, table=True):
    __tablename__ = "granted_user_role"

    user_id: UUID = Field(default=None, primary_key=True, foreign_key="user.id")
    role_id: UUID = Field(default=None, primary_key=True, foreign_key="role.id")


# granted_user_role_table = Table(
#     "granted_user_role",
#     Base.metadata,
#     Column("user_id", ForeignKey("user.id"), primary_key=True),
#     Column("role_id", ForeignKey("role.id"), primary_key=True),
#     UniqueConstraint(  # Don't allow entries with the same account/role combination (no duplicate entries)
#         "user_uuid", "role_uuid", name="_granted_user_role_uc"
#     ),
# )


# --------------------
# -- Entity Tables --
# --------------------
# Lookup table
# A user role which can be assigned to a user to grant associated
# permissions.
class RoleTbl(SQLModel, table=True):
    __tablename__ = "role"

    id: UUID = Field(
        primary_key=True,
        default_factory=uuid4,
        sa_column_kwargs={
            "server_default": text("gen_random_uuid()"),
        },
    )
    name: str = Field(max_length=20, unique=True)
    description: str = Field(max_length=256, unique=False, nullable=True)
    permissions: list["PermissionTbl"] = Relationship(
        link_model=RolePermissionLinkTbl, back_populates="user_roles"
    )


# Lookup table
# Specific permissions which are checked for to perform an action
class PermissionTbl(SQLModel, table=True):
    __tablename__ = "permission"

    id: UUID = Field(
        primary_key=True,
        default_factory=uuid4,
        sa_column_kwargs={
            "server_default": text("gen_random_uuid()"),
        },
    )
    name: str = Field(max_length=30, unique=False)
    description: str = Field(max_length=256, nullable=True)
    user_roles: list["RoleTbl"] = Relationship(
        link_model=RolePermissionLinkTbl, back_populates="permissions"
    )



import logging

from sqlalchemy import select, and_, desc
from sqlmodel import Session

# from sqlalchemy.sql import values
from datetime import datetime

from ceonmedia_db.models.job_models import (
    JobTbl,
    JobProgressTbl,
)

# from db.engine import SessionLocal

logger = logging.getLogger(__name__)


class Job:
    @classmethod
    def get(cls, db: Session, job_uuid):
        stmt = select(JobTbl).where(JobTbl.uuid == job_uuid)
        got_user = db.execute(stmt).scalars().one()
        return got_user

    @classmethod
    def get_all(cls, db: Session, page=0, limit=15):
        skip = page * limit
        stmt = select(JobTbl).offset(skip).limit(limit).order_by(desc(JobTbl.uuid))
        result = db.execute(stmt).scalars().all()
        return result

    @classmethod
    def get_jobs_by_user_id(cls, db: Session, user_uuid, page=0, limit=15):
        skip = page * limit
        stmt = (
            select(JobTbl)
            .where(JobTbl.user_uuid == user_uuid)
            .offset(skip)
            .limit(limit)
            .order_by(desc(JobTbl.created_at))
        )
        result = db.execute(stmt).scalars().all()
        return result

    # TODO refactor above func so filtering (see below func) can be an optional argument

    @classmethod
    def get_jobs_by_job_id_list(
        cls, db: Session, user_uuid, job_id_list, page=0, limit=15
    ):
        """returns a user's jobs, filtered by a list of job ids"""
        skip = page * limit
        stmt = (
            select(JobTbl)
            # .where(JobTbl.user_uuid == user_uuid)
            .where(
                and_(
                    JobTbl.user_uuid == user_uuid,
                    JobTbl.uuid.in_(job_id_list),
                )
            )
            .offset(skip)
            .limit(limit)
        )
        result = db.execute(stmt).scalars().all()
        return result

    @classmethod
    def create(cls, db: Session, **values):
        # Create new job entry
        new_job = JobTbl(**values)
        db.add(new_job)
        db.flush()
        logger.debug(f"Created new_job: {new_job}")
        # Create new job_progress entry
        new_job_progress = JobProgressTbl(job_uuid=new_job.uuid)
        db.add(new_job_progress)
        logger.debug(f"Created new_job_progress: {new_job_progress}")
        db.commit()

        return new_job


class JobProgress:
    @classmethod
    def get(cls, db: Session, job_uuid):
        stmt = select(JobProgressTbl).where(JobProgressTbl.job_uuid == job_uuid)
        got_job = db.execute(stmt).scalars().one()
        return got_job

    @classmethod
    def get_all(cls, db: Session):
        # TODO only return jobs belonging to user id
        stmt = select(JobProgressTbl).order_by(desc(JobProgressTbl.job_uuid))
        result = db.execute(stmt).scalars().all()
        return result

    @classmethod
    def create(cls, db: Session, **values):
        new_job_progress = JobProgressTbl(**values)
        db.add(new_job_progress)
        db.commit()
        return new_job_progress

    @classmethod
    def set_timestamp(cls, db: Session, job_uuid, columns_to_timestamp):
        """
        job_uuid: The id of the job to update
        columns_to_timestamp: A list of strings that are the column names
        """
        if not columns_to_timestamp:
            return None
        if type(columns_to_timestamp) is not list:
            columns_to_timestamp = [columns_to_timestamp]

        timestamp = datetime.now()
        msgs_to_print = ["Got keys to set: "]
        values_to_set = {}
        for key in columns_to_timestamp:
            msgs_to_print.append(f"\n\t\t{key}")
            values_to_set[key] = timestamp
        logger.info("".join(msgs_to_print))
        stmt = (
            JobProgressTbl.update()
            .where(JobProgressTbl.uuid == job_uuid)
            .values(**values_to_set)
        )
        try:
            result = db.execute(stmt).scalars().one()
            logger.info(f"Succesfully set timestamps: {values_to_set}")
            return values_to_set
        except Exception as e:
            logger.warn(f"Failed to set timestamp in jobqueue: {e}")
            raise Exception(f"Failed to update jobqueue timestamps: {values_to_set}")



import logging
import pytest

from sqlmodel import Session

from ceonmedia_db import crud
from ceonmedia_db import models
from ceonmedia_db import schemas
from ceonmedia_db import db_setup
from ceonmedia_db.tests import TestSessionLocal

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel("INFO")

# TODO split this module into three test modules
# test_user_create
# test_user_update
# test_user_get (e.g. get by role (admin, affiliate) etc)


def add_user(db: Session, email: str, password: str) -> models.UserAccountTbl:
    user_to_create = schemas.user_schemas.UserAccountCreate(
        email=email, password=password
    )
    logger.info(f"Creating user: {user_to_create.email}")
    user_in_db = crud.user_crud.create(db, user_to_create)
    logger.debug(f"Successfully added user: {user_in_db}")
    return user_in_db


def delete_user(db: Session, User):
    raise NotImplementedError


# TODO test_get_user():
# TODO test_get_user_by_email():
# TODO test_delete_user():
# TODO test_add_user_fail_already_exists():
# TODO test_add_user_with_role_success():


@pytest.fixture
def existing_user():
    return models.UserAccountTbl(email="existing_user@gmail.com")


@pytest.fixture
def db_session():
    return TestSessionLocal()


# @pytest.fixture
# def empty_db_session(db_session: Session):
# """Reset the database to the testing state"""
# with db_session as session:
# db_setup.reset.reset_schema(session)
# return db_session


# @pytest.fixture(autouse=True)
@pytest.fixture
def default_db_session(db_session: Session):
    """Reset the database to the testing state"""
    with db_session as session:
        db_setup.reset.reset_schema(session)
        db_setup.users.create_default_users(session)
        # TODO setup default testing data
    return db_session


def test_add_user_success(default_db_session: Session):
    with default_db_session as session:
        # reset_schema(session)
        db_setup.reset.reset_schema(session)
        user_from_db = add_user(session, "added_user@hotmail.com", "123")
        add_user(session, "testuser3@hotmail.com", "123")
        got_user = crud.user_crud.get_by_email(session, user_from_db.email)
        logger.info(f"Read user: {got_user}")


def test_get_user_success(
    default_db_session: Session, existing_user: models.UserAccountTbl
):
    with default_db_session as session:
        got_user = crud.user_crud.get_by_email(session, existing_user.email)
    logger.info(f"Read got user: {got_user}")
    assert got_user != None



# Pydantic models, renamed to 'schemas' to avoid being confused with sqlalchemy models.
from typing import Optional  # Extends data type inputs for fastapi
from pydantic import ConfigDict, BaseModel
from uuid import UUID


class Token(BaseModel):
    access_token: str
    token_type: str


class TokenData(BaseModel):
    username: Optional[str] = None


class UserBase(BaseModel):
    email: str


# Data required to create a user
class UserCreate(UserBase):
    password: str
    first_name: Optional[str] = None
    last_name: Optional[str] = None


# The user in the DB
class User(UserBase):
    uuid: UUID
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    avatar: Optional[str] = None
    model_config = ConfigDict(from_attributes=True)


class ProjectCreate(BaseModel):
    title: str
    description: Optional[str] = None
    proj_inputs: dict


class JobBase(BaseModel):
    job_id: int


class JobCreate(JobBase):
    user_id: int
    project_id: int


# class Job(JobBase):



from uuid import uuid4
from typing import Optional

from sqlalchemy import (
    String,
    Boolean,
    DateTime,
    ForeignKey,
    func,
    text,
)
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship
from sqlalchemy.dialects.postgresql import UUID

from ceonmedia_db.models.base import Base
from . import access_control


# Parent table for the user entity
class UserAccountTbl(Base):
    __tablename__ = "user_account"

    uuid: Mapped[UUID] = mapped_column(
        UUID(as_uuid=True),
        primary_key=True,
        nullable=False,
        index=True,
        unique=True,
        default=uuid4,
        server_default=text("gen_random_uuid()"),
    )
    email: Mapped[str] = mapped_column(String(255), unique=True, index=True)
    email_is_confirmed: Mapped[bool] = mapped_column(Boolean, server_default="f")
    created_at: Mapped[DateTime] = mapped_column(
        DateTime(timezone=True), server_default=func.now()
    )
    updated_at: Mapped[DateTime] = mapped_column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
    )

    # Relationships
    user_profile: Mapped["UserProfileTbl"] = relationship(back_populates="user")
    user_projects: Mapped[list["ProjectTbl"]] = relationship(back_populates="owner")
    user_roles: Mapped[list["UserRoleTbl"]] = relationship(  # type: ignore
        secondary=access_control.granted_user_role_table
    )
    user_login: Mapped["UserLoginTbl"] = relationship(back_populates="user")
    reusable_preview_credits: Mapped[list["ReusablePreviewTokenTbl"]] = relationship(
        back_populates="user_account"
    )

    def __repr__(self):
        return f"User(uuid={self.uuid!r}, email={self.email!r})"


# ------------------
# -- Child tables --
# ------------------


class UserLoginTbl(Base):
    __tablename__ = "user_login"

    user_account_uuid: Mapped[UUID] = mapped_column(
        UUID(as_uuid=True),
        ForeignKey("user_account.uuid", ondelete="CASCADE"),
        primary_key=True,
        unique=True,
    )
    hashed_password: Mapped[str] = mapped_column(String(60))
    password_recovery_token: Mapped[Optional[UUID]] = mapped_column(
        UUID(as_uuid=True),
        unique=True,
    )
    user: Mapped["UserAccountTbl"] = relationship(back_populates="user_login")

    # In this case we don't want to add user_login to UserTbl,
    # so don't back_populate?
    # user: Mapped["UserAccountTbl"] = relationship()
    # user: Mapped["UserTbl"] = relationship(back_populates="user_login")


class UserProfileTbl(Base):
    __tablename__ = "user_profile"

    user_account_uuid: Mapped[UUID] = mapped_column(
        UUID(as_uuid=True),
        ForeignKey("user_account.uuid", ondelete="CASCADE"),
        primary_key=True,
        unique=True,
    )
    username: Mapped[Optional[str]] = mapped_column(String(20), unique=True)
    avatar: Mapped[Optional[str]] = mapped_column(String(255))
    first_name: Mapped[Optional[str]] = mapped_column(String(20))
    last_name: Mapped[Optional[str]] = mapped_column(String(20))

    user: Mapped["UserAccountTbl"] = relationship(back_populates="user_profile")
    # DoB? How-to Date (not DateTime)
    # WebSite?
    # IMDB?



import logging


from ceonmedia_db import crud
from ceonmedia_db import schemas

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel("INFO")


NEW_USER_CREATE = schemas.UserAccountCreate(
    email="newly_created_user@hotmail.com", password="123"
)


# Create a simple, regular user
def test_token_create_success(default_db_session):
    with default_db_session() as session:
        # reset_schema(session)
        # TODO get test user(s) from... a 'fixutre'?
        user = crud.user.get_by_email(session, "admin@ceonmedia.com")
        logger.info(f"Got user for testing email verification tokens: {user.email}")

        token = crud.security.create_email_verification_token(
            session, str(user.uuid), user.email
        )
        logger.info(f"Created token: {token.__dict__}")
        assert token.user_account_uuid == user.uuid


# TODO add test for email changing (token redemption process is hte same, but a token must be
# created that sets the email to a value different from the current email)


def test_token_redeem_success(default_db_session):
    # This test assumes handling a newly registered user that has not yet confirmed their email
    with default_db_session() as session:
        # TODO get test user(s) from... a 'fixutre'?
        user = crud.user.get_by_email(session, "admin@ceonmedia.com")
        user.email_is_confirmed = False
        logger.info(
            f"Got user for testing email verification tokens: {user.email}, email_is_confirmed: {user.email_is_confirmed}"
        )

        # Create a token for the user
        token = crud.security.create_email_verification_token(
            session, str(user.uuid), user.email
        )
        logger.info(f"Created token: {token.__dict__}")
        assert token.user_account_uuid == user.uuid

        # Redeem the token
        logger.info(f"Redeeming email verification token...")
        crud.security.redeem_email_verification_token(session, str(token.uuid))

        # Confirm that user's email was updated to match the email in the token
        # And that email_is_confirmed is True
        assert user.email == token.email_to_verify
        assert user.email_is_confirmed == True



# Pydantic models, renamed to 'schemas' to avoid being confused with sqlalchemy models.
from typing import Optional

from pydantic import BaseModel


class UserProfileCreate(BaseModel):
    username: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    avatar: Optional[str] = None


# Data required to create a new user
class UserAccountCreate(BaseModel):
    email: str
    password: str
    profile: Optional[UserProfileCreate] = None


# WIP move all non-create schemas to the API/external applications
# Only handle DB creation schemas/data in this package

# class UserAccountOut(UserAccountBase):
#   profile: Optional[UserProfileOut]

# ORM mode for the class that mirrors the DB
# class Config:
#   orm_mode = True



from . import security
from . import oauth_account
from . import user
from . import user_credit
from . import stripe
from . import access_control
from . import project
from . import submission
from . import job



from uuid import uuid4, UUID
from datetime import date, timedelta, datetime, timezone

from sqlmodel import SQLModel, Field, TIMESTAMP

from sqlalchemy import (
    func,
    text,
)


def get_valid_until():
    valid_until = date.today() + timedelta(days=30)
    return valid_until


class JobTbl(SQLModel, table=True):
    __tablename__ = "job"

    id: UUID = Field(
        primary_key=True,
        default_factory=uuid4,
        index=True,
        sa_column_kwargs={
            "server_default": text("gen_random_uuid()"),
        },
    )
    user_account_id: UUID = Field(foreign_key="user.id", ondelete="CASCADE")
    project_id: UUID = Field(foreign_key="project.id", ondelete="CASCADE")
    submission_id: UUID = Field(
        # UUID(as_uuid=True),
        foreign_key="submission.id",
        ondelete="SET NULL",
        nullable=True,
    )
    # TODO allow type checking to enforce enum?
    job_type: str = Field(max_length=30, nullable=False)
    # job_type: Mapped[CstockJob.job_types] = mapped_column(max_length=30, nullable=False)
    # foreign_key="submission.id", ondelete="SET NULL"),
    created_at: datetime = Field(
        default_factory=lambda: datetime.now(timezone.utc),
        nullable=False,
        sa_column_kwargs={"server_default": func.now()},
    )
    updated_at: datetime | None = Field(
        default_factory=lambda: datetime.now(timezone.utc),
        sa_column_kwargs={
            "onupdate": func.now(),
            "server_default": func.now(),
        },
        sa_type=TIMESTAMP(timezone=True),
    )
    ended_at: datetime = Field(default=None, nullable=True)
    success: bool = Field(default=None, nullable=True)

    # project: "ProjectTbl" = Relationship()
    # user_account: "UserTbl" = Relationship()


class JobProgressTbl(SQLModel, table=True):
    __tablename__ = "job_progress"

    job_id: UUID = Field(
        foreign_key="job.id",
        ondelete="CASCADE",
        primary_key=True,
    )
    assigned_to_node: str = Field(nullable=True)
    assigned_at: datetime = Field(nullable=True)
    started_at: datetime = Field(nullable=True)
    create_job_inputs_start: datetime = Field(nullable=True)
    create_job_inputs_end: datetime = Field(nullable=True)
    prepare_files_start: datetime = Field(nullable=True)
    prepare_files_end: datetime = Field(nullable=True)
    render_submit_start: datetime = Field(nullable=True)
    render_submit_end: datetime = Field(nullable=True)
    wait_for_render_start: datetime = Field(nullable=True)
    wait_for_render_end: datetime = Field(nullable=True)
    upload_outputs_start: datetime = Field(nullable=True)
    upload_outputs_end: datetime = Field(nullable=True)
    updated_at: datetime = Field(
        sa_column_kwargs={
            "onupdate": func.now(),
            "server_default": func.now(),
        },
    )


# TODO setup this by converting TIMESTAMP objects to strings then to datetime objects?
"""
job_progress_tbl.d.latest_timestamp = (
    func.max(
        [
            job_progress_tbl.c.started_at,
            job_progress_tbl.c.render_submit_start,
            job_progress_tbl.c.render_complete_confirmed,
            job_progress_tbl.c.upload_outputs_start,
            job_progress_tbl.c.ended_at,
        ]
    )
).label("latest_timestamp")
"""
# Can be late accessed as job_progress_tbl.d.latest_timestamp as though it were a real column



import logging
from sqlalchemy import exc
from psycopg2 import errors as psy_errors

logger = logging.getLogger(__name__)


class CstockDBError(Exception):
    pass


class NotFoundInDBError(CstockDBError):
    pass


class AlreadyExistsInDBError(CstockDBError):
    pass


def handle_sql_error(sql_error, msg=None):
    """Receives an SQL error and raises the appropriate custom error"""
    logger.debug("Handling error...")
    if isinstance(sql_error, exc.IntegrityError):
        logger.debug("Got IntegrityError...")
        _handle_sql_integrity_error(sql_error, msg=msg)
    raise sql_error


def _handle_sql_integrity_error(sql_integrity_error: exc.IntegrityError, msg=None):
    if isinstance(sql_integrity_error.orig, psy_errors.UniqueViolation):
        logger.debug("Got UniqueViolation...")
        default_msg = f"UniqueViolation: {sql_integrity_error}"
        message_to_raise = msg if msg else default_msg
        raise AlreadyExistsInDBError(message_to_raise)
    if isinstance(sql_integrity_error.orig, psy_errors.ForeignKeyViolation):
        default_msg = f"ForeignKeyViolation: {sql_integrity_error}"
        message_to_raise = msg if msg else default_msg
        logger.debug("Got ForeignKeyViolation...")
        raise NotFoundInDBError(message_to_raise)
    raise Exception(f"Unknown SQL Integrity error :O : {sql_integrity_error}")



# For handling user tokens for free project renders, free previews etc.
# Also acts as a way to give a user a free purchase of a project.
import logging
from datetime import datetime, timezone, timedelta
from uuid import UUID

from sqlmodel import Session
from sqlalchemy import exc
from sqlalchemy import select

from ceonmedia_db import models
from ceonmedia_db import errors
from ceonmedia_db import constants

logger = logging.getLogger(constants.LOGGER_NAME)

# def get_by_email(db: Session, email: str) -> user_models.UserAccountTbl:
#     stmt = select(user_models.UserAccountTbl).where(
#         user_models.UserAccountTbl.email == email
#     )
#     try:
#         got_user = db.execute(stmt).scalars().one()
#     except exc.NoResultFound:
#         raise errors.NotFoundInDBError(f"Could not find user by email: {email}")
#     logger.debug(f"Returning: {got_user}")
#     return got_user


# def get_all(db: Session):
#     stmt = select(models.user_credit.ReusablePreviewTokenTbl).order_by(
#         models.user_credit.ReusablePreviewTokenTbl.created_at
#     )
#     result = db.execute(stmt).scalars().all()
#     logger.debug(f"result: {result}")
#     return result


# TODO get by comparing available date
# def get_available_reusable_credits_for_user(db: Session, user_uuid: UUID):
#     stmt = select(models.user_credit.ReusablePreviewTokenTbl).where(
#         models.user_credit.ReusablePreviewTokenTbl.user_account_uuid == user_uuid
#     )
#     try:
#         credits_in_db = db.execute(stmt).scalars().all()
#     except exc.NoResultFound:
#         raise errors.NotFoundInDBError(
#             f"Could not find credits for user uuid: {user_uuid}"
#         )
#     logger.debug(f"Returning: {credits_in_db}")
#     return credits_in_db


def get_reusable_credits_for_user(db: Session, user_id: UUID):
    stmt = select(models.user_credit.ReusablePreviewTokenTbl).where(
        models.user_credit.ReusablePreviewTokenTbl.user_account_id == user_id
    )
    try:
        credits_in_db = db.execute(stmt).scalars().all()
    except exc.NoResultFound:
        raise errors.NotFoundInDBError(f"Could not find credits for user id: {user_id}")
    logger.debug(f"Returning: {credits_in_db}")
    return credits_in_db


# def redeem_reusable_preview_credit(db: Session, reusable_preview_credit_id: UUID):
#     def is_available(preview_credit_in_db):
#         return preview_credit_in_db.available_at <= datetime.now(timezone.utc)

#     def calculate_next_available_time(wait_time_days: float = 30):
#         return datetime.now(timezone.utc) + timedelta(days=wait_time_days)

#     stmt = select(models.user_credit.ReusablePreviewTokenTbl).where(
#         models.user_credit.ReusablePreviewTokenTbl.id == reusable_preview_credit_id
#     )

#     try:
#         reusable_preview_credit_in_db = db.execute(stmt).scalar_one()
#     except exc.NoResultFound:
#         raise errors.NotFoundInDBError(
#             f"Could not find reusable preview credit with id: {reusable_preview_credit_id}"
#         )
#     logger.info(f"Got reusable credit for redemption: {reusable_preview_credit_in_db}")
#     available_at = reusable_preview_credit_in_db.available_at
#     logger.info(f"Available at: {available_at}")
#     if not is_available(reusable_preview_credit_in_db):
#         raise Exception(
#             "Tried to redeem an unavailable render credit TODO proper exception handling"
#         )
#     next_available_time = calculate_next_available_time()
#     logger.info(f"Setting new available_at: {next_available_time}")
#     reusable_preview_credit_in_db.available_at = next_available_time
#     logger.warning("Commiting change to db...")
#     db.commit()

#     logger.warning(f"Returning: {reusable_preview_credit_in_db}")
#     return reusable_preview_credit_in_db


def redeem_reusable_preview_credit_for_user(db: Session, user_id: UUID):
    # Get the first available reusable preview credit and redeem it (update the 'available at' to a future date)
    def is_available(preview_credit_in_db):
        return preview_credit_in_db.available_at <= datetime.now(timezone.utc)

    def calculate_next_available_time(wait_time_days: float = 30):
        return datetime.now(timezone.utc) + timedelta(days=wait_time_days)

    stmt = select(models.user_credit.ReusablePreviewTokenTbl).where(
        models.user_credit.ReusablePreviewTokenTbl.user_account_id == user_id,
        models.user_credit.ReusablePreviewTokenTbl.available_at
        <= datetime.now(timezone.utc),
    )

    try:
        reusable_preview_credit_in_db = db.scalars(stmt).first()
        if not reusable_preview_credit_in_db:
            raise errors.NotFoundInDBError(
                f"Could not find available preview credit for user with id: {user_id}"
            )
    except exc.NoResultFound:
        raise errors.NotFoundInDBError(
            f"Could not find available preview credit for user with id: {user_id}"
        )
    logger.info(f"Got reusable credit for redemption: {reusable_preview_credit_in_db}")
    available_at = reusable_preview_credit_in_db.available_at
    logger.info(f"Available at: {available_at}")
    if not is_available(reusable_preview_credit_in_db):
        raise Exception(
            "Tried to redeem an unavailable render credit TODO proper exception handling"
        )
    next_available_time = calculate_next_available_time()
    logger.info(f"Setting new available_at: {next_available_time}")
    reusable_preview_credit_in_db.available_at = next_available_time
    logger.warning("Commiting change to db...")
    db.commit()
    logger.warning("Committed.")

    logger.warning(f"Returning: {reusable_preview_credit_in_db}")
    return reusable_preview_credit_in_db


def create_reusable_preview_credit(db: Session, user_account_id: UUID):
    logger.debug(f"Creating new credit for user uuid: {user_account_id}")
    new_credit = models.user_credit.ReusablePreviewTokenTbl(
        user_account_id=user_account_id
    )
    logger.debug(f"Adding new token: {new_credit}")
    db.add(new_credit)
    logger.debug(f"Added new credit to session: {new_credit}")

    logger.debug(f"Flushing new credit to session: {new_credit}")
    try:
        db.flush()
    except exc.SQLAlchemyError as e:
        msg = f"Failed to create new credit for user id: {user_account_id}"
        errors.handle_sql_error(e, msg=msg)
    logger.debug(f"Flushed new credit to session: {new_credit}")

    # -- Commit Changes --
    try:
        db.commit()
    except exc.SQLAlchemyError as e:
        msg = f"Failed to commit changes for new credit for user_id: {user_account_id}"
        errors.handle_sql_error(e, msg=msg)
    logger.info(f"Committed new credit to db: {new_credit.id}")

    return new_credit



import logging

from sqlmodel import Session

from ceonmedia_db import crud
from ceonmedia_db import schemas

logger = logging.getLogger(__name__)

LOCAL_PROJECTS_DIR = (
    "/mnt/FileStorage/Dayne/Web/proj_ceonmedia/assets/my_cstock_projects"
)
PROJECTS_TO_GET = ["papers_on_desk", "Waving_Flag"]


def get_local_projects():
    pass


def push_local_projects_to_db(
    session: Session,
    projects_dir=LOCAL_PROJECTS_DIR,
    projects_to_get: list[str] = PROJECTS_TO_GET,
):
    logger.info("Pushing default projects (from local filesystem) to db...")
    # TODO actually get projects
    owner = crud.user.get_by_email(session, "admin@ceonmedia.com")
    project_to_create = schemas.ProjectCreate(
        title="Existing Project",
        description="This project has existed since the start",
        proj_inputs=[{"test": "abc"}],
        owner_id=owner.id,
    )
    logger.info(f"Creating project; {project_to_create}")
    project_in_db = crud.project.create(session, project_to_create)
    logger.info(f"Created project in db: {project_in_db}")
    return project_in_db


def push_to_db(session: Session):
    owner = crud.user.get_by_email(session, "admin@ceonmedia.com")
    project_to_create = schemas.ProjectCreate(
        title="Existing Project",
        description="This project has existed since the start",
        proj_inputs=[{"test": "abc"}],
        owner_id=owner.id,
    )
    logger.info(f"Creating project; {project_to_create}")
    project_in_db = crud.project.create(session, project_to_create)
    logger.info(f"Created project in db: {project_in_db}")
    return project_in_db



# Pydantic models, renamed to 'schemas' to avoid being confused with sqlalchemy models.
from uuid import UUID

# from pydantic import BaseModel, Field, EmailStr


class JobBase:
    job_uuid: UUID


class JobCreate(JobBase):
    user_uuid: UUID
    project_uuid: UUID


# class Job(JobBase):

# class Config:
#     orm_mode = True



"""Baseline

Revision ID: 4cec02cae1c1
Revises:
Create Date: 2023-11-18 15:50:35.279991

"""

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "4cec02cae1c1"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "user_stripe",
        sa.Column("user_account_uuid", sa.UUID(), nullable=False),
        sa.Column("stripe_customer_id", sa.String(length=30), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_account_uuid"], ["user_account.uuid"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("user_account_uuid"),
        sa.UniqueConstraint("stripe_customer_id"),
        sa.UniqueConstraint("user_account_uuid"),
    )
    op.create_unique_constraint(
        "_granted_permission_uc",
        "granted_permission",
        ["user_role_uuid", "permission_uuid"],
    )
    op.create_unique_constraint(
        "_granted_user_role_uc",
        "granted_user_role",
        ["user_account_uuid", "user_role_uuid"],
    )
    op.create_unique_constraint(None, "user_login", ["user_account_uuid"])
    op.create_unique_constraint(None, "user_profile", ["user_account_uuid"])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "user_profile", type_="unique")
    op.drop_constraint(None, "user_login", type_="unique")
    op.drop_constraint("_granted_user_role_uc", "granted_user_role", type_="unique")
    op.drop_constraint("_granted_permission_uc", "granted_permission", type_="unique")
    op.drop_table("user_stripe")
    # ### end Alembic commands ###



import pytest
import logging
from ceonmedia_db.tests import TestSessionLocal
from ceonmedia_db import db_setup

# fixtures defined here in conftest.py can be used by the tests in any modules

logger = logging.getLogger(__name__)


# Making TestSessionLocal accessible via fixture so we don't have to
# manually import it to every test module
@pytest.fixture(scope="session")
def db_session():
    """Provides the database connection to be used for all testing.
    The schema is reset and the DB is empty by default.
    Use the other fixtures to provide testing data"""
    with TestSessionLocal() as session:
        engine = session.get_bind()
        logger.warning(f"Got session bind: {engine}")
        logger.warning(f"Got session engine uri: {engine.url}")
        db_setup.reset.reset_schema(session)
    return TestSessionLocal


@pytest.fixture()
def default_db_session(db_session):
    """Delete existing DB entries and reset the testing state"""
    with db_session() as session:
        # TODO delete/reset entries from other tests?
        # Refactor db_setup/testing and a master db_setup.testing.insert_testing_data
        logger.warning("Fixture resetting DB entries...")
        db_setup.reset.clear_data(session)
        db_setup.access_control.push_to_db(session)
        db_setup.users.create_default_users(session)
        db_setup.project.push_to_db(session)
    return db_session


@pytest.fixture()
def empty_db_session(db_session):
    """Provides an empty DB with only the schema initialized"""
    with db_session() as session:
        logger.warning("Fixture clearing data")
        db_setup.reset.clear_data(session)
        # Refactor db_setup/testing and a master db_setup.testing.insert_testing_data
    return db_session



from logging.config import fileConfig
import os, sys
from dotenv import load_dotenv

# from sqlalchemy import engine_from_config
from sqlalchemy import create_engine

from alembic import context

alembic_dir = os.path.dirname(os.path.realpath(__file__))
# logger.info(f"Got alembic_dir: {alembic_dir}")
sys.path.append(alembic_dir)
load_dotenv(os.path.join(alembic_dir, ".env"))


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from ceonmedia_db.database import Base

# from ceonmedia_db.models.user_models import *
# from ceonmedia_db.models.project_models import *
# from ceonmedia_db.models import *
from ceonmedia_db.models.base import Base

target_metadata = Base.metadata
# target_metadata = None


# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def get_url():
    DB_URI = os.getenv("DB_URI")
    # logger.info(f"Alembic got DB_URI: {DB_URI}")
    if not DB_URI:
        raise Exception("Environment variable DB_URI not set")
    return DB_URI


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    # url = config.get_main_option("sqlalchemy.url")
    url = get_url()
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    # connectable = engine_from_config(
    # config.get_section(config.config_ini_section),
    # prefix="sqlalchemy.",
    # poolclass=pool.NullPool,
    # )
    connectable = create_engine(get_url())

    with connectable.connect() as connection:
        context.configure(connection=connection, target_metadata=target_metadata)

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



import logging
from uuid import UUID

from sqlmodel import Session
from sqlalchemy import exc
from sqlalchemy import select

from ceonmedia_db import models
from ceonmedia_db import errors
from ceonmedia_db import constants


logger = logging.getLogger(constants.LOGGER_NAME)


def get_customer(db: Session, user_id: UUID) -> models.stripe.UserStripeTbl:
    stmt = select(models.stripe.UserStripeTbl).where(
        models.stripe.UserStripeTbl.user_account_id == user_id
    )
    try:
        got_user = db.execute(stmt).scalars().one()
    except exc.NoResultFound:
        raise errors.NotFoundInDBError(f"Could not find user by id: {user_id}")
    logger.debug(f"Returning: {got_user}")
    return got_user


def create_customer(
    db: Session, user_id: UUID, stripe_customer_id: str
) -> models.stripe.UserStripeTbl:
    logger.debug(f"Creating new stripe id: {stripe_customer_id} for user {user_id}")
    new_stripe_customer = models.stripe.UserStripeTbl(
        user_account_id=user_id, stripe_customer_id=stripe_customer_id
    )
    logger.debug(f"Saving new stripe user: {new_stripe_customer}")
    db.add(new_stripe_customer)
    logger.debug(f"Added new stripe user to session: {new_stripe_customer}")

    # -- Commit Changes --
    try:
        db.commit()
    except exc.SQLAlchemyError as e:
        msg = (
            f"Failed to commit changes (adding stripe customer) for new user: {user_id}"
        )
        errors.handle_sql_error(e, msg=msg)
    logger.info(
        f"Committed new stripe customer to db: {new_stripe_customer.stripe_customer_id}"
    )

    return new_stripe_customer



# Authentication
import time
from typing import Dict

from passlib.context import CryptContext
from jose import jwt

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")

# TODO user environment vars (investigate decouple/config)
# from decouple import config
# JWT_SECRET = config("secret")
# JWT_ALGORITHM = config("algorithm")

# to get a string like this run:
# openssl rand -hex 32
JWT_SECRET = "cb1ccdcb5cc537b73106ff69f87e7c218c56e7318d1028c149ff8b82a5c4689f"
JWT_ALGORITHM = "HS256"
ACCESS_TOKEN_EXPIRE_MINUTES = 60


def verify_password(plain_password, hashed_password):
    return pwd_context.verify(plain_password, hashed_password)


def get_password_hash(password):
    return pwd_context.hash(password)


def token_response(token: str):
    return {"access_token": token}


def signJWT(user_id: str) -> Dict[str, str]:
    payload = {"user_id": user_id, "expires": time.time() + (60 * 60)}
    token = jwt.encode(payload, JWT_SECRET, algorithm=JWT_ALGORITHM)

    return token_response(token)


def decodeJWT(token: str) -> dict:
    try:
        decoded_token = jwt.decode(token, JWT_SECRET, algorithms=[JWT_ALGORITHM])
        return decoded_token if decoded_token["expires"] >= time.time() else None
    except:
        return {}



import logging
from sqlmodel import Session
from ceonmedia_db import crud

from ceonmedia_db.schemas import UserRoleOption
from ceonmedia_db.schemas import PermissionOption

logger = logging.getLogger(__name__)
logger.setLevel("INFO")


GRANTED_PERMISSIONS = {
    UserRoleOption.VERIFIED_USER: [
        PermissionOption.SUBMIT_PREVIEWS,
        PermissionOption.PURCHASE,
    ],
    UserRoleOption.ADMIN: [permission for permission in PermissionOption],
}


def create_user_roles(session: Session):
    for user_role in UserRoleOption:
        logger.info(f"\tCreating user_role '{user_role}'")
        crud.access_control.create_role(session, user_role)


def create_permissions(session: Session):
    for permission in PermissionOption:
        logger.info(f"\tCreating permission '{permission}'")
        crud.access_control.create_permission(session, permission)


def assign_permissions_to_roles(session: Session):
    for role, permissions in GRANTED_PERMISSIONS.items():
        logger.info(f"\tCreating permission assignments for role '{role}':")
        for permission in permissions:
            crud.access_control.add_permission_to_role(session, role, permission)


def print_permissions(session):
    permissions_in_db = crud.access_control.get_all_permissions(session)
    logger.info(
        f"Found permissions in DB: {[permission_in_db.value for permission_in_db in permissions_in_db]}"
    )


def print_user_roles(session):
    roles_in_db = crud.access_control.get_all_user_roles(session)
    logger.info(
        f"Found roles in DB: {[role_in_db.value for role_in_db in roles_in_db]}"
    )


def push_to_db(session: Session):
    # TODO a way to UPDATE permissions, without a full DB reset.
    logger.info(GRANTED_PERMISSIONS)
    print()
    logger.info("-- Creating user_roles in db... --")
    create_user_roles(session)
    print()
    logger.info("-- Creating permission in db... --")
    create_permissions(session)

    # Assign permissions
    print()
    # TODO
    logger.info("-- Assigning permissions to roles in db... --")
    assign_permissions_to_roles(session)



import os
import logging
import dotenv

import db_setup_scripts

from sqlmodel import create_engine, Session

dotenv.load_dotenv("./.env")

root_logger = logging.getLogger()
root_logger.setLevel("DEBUG")

logger = logging.getLogger("db_scripts")

DEFAULT_CONNECT_ARGS = {"sslmode": "require"}
connect_args_local_docker = {}


def create_session(db_uri: str, connect_args=None) -> Session:
    if connect_args == None:
        connect_args = DEFAULT_CONNECT_ARGS
    logger.warning(
        "TODO make sure to use connect_args with sslmode require in production"
    )
    logger.warning(f"{connect_args=}")
    engine = create_engine(
        # DB_URI, connect_args=connect_args, pool_pre_ping=True
        db_uri,
        connect_args=connect_args,
        pool_pre_ping=True,
    )
    return Session(engine)


def main():
    DB_URI = os.getenv("DB_URI")
    print(f"Got DB_URI: {DB_URI}")
    if not DB_URI:
        raise Exception("Did not receive a DB_URI")
    # For local dev, disable ssl require
    print(f"WARNING: You are about to reset the DB at:")
    print(f"\t{DB_URI}")

    if not confirm_user_intent("Continue?"):
        print("Exiting.")
        return

    connect_args_dev = {}
    with create_session(DB_URI, connect_args=connect_args_dev) as session:
        db_setup_scripts.reset.reset_schema(session)
        print(f"Database has been reset and rebuilt with the latest schema.")

        print(f"Populating default data (user roles)")
        # Populate with fundamental/required data such as user roles
        setup_default_state(session)

        # Populate with default data
        # if confirm_user_intent(f"Populate database with default data?"):
        #     setup_default_entries(session)


def setup_default_state(session):
    """
    State which is assumed to always be desired as setup
        - Access control (user roles and assigned permissions)
    """
    db_setup_scripts.access_control.push_to_db(session)
    db_setup_scripts.products.create_default_products(session)


def setup_default_entries(session):
    """
    Default database entries that may or may not be desired.
        - Users
        - Projects
    """
    db_setup_scripts.users.create_default_users(session)


def confirm_user_intent(msg: str):
    user_resp = input(f"{msg} ('yes' to continue): ")
    if user_resp.lower() == "yes":
        return True
    return False


if __name__ == "__main__":
    main()



import logging
from typing import Optional
from sqlmodel import Session
from ceonmedia_db import schemas
from ceonmedia_db import crud
from ceonmedia_db.schemas import UserRoleOption

logger = logging.getLogger(__name__)


class DefaultUser:
    def __init__(
        self,
        user_create: schemas.UserAccountCreate,
        roles: Optional[list[UserRoleOption]] = None,
    ):
        self.user_account = user_create
        self.roles = roles


# TODO for a more consistent workflow, replace with pydantic creation classes similar to
# what the API user will be working with.
DEFAULT_USERS = [
    DefaultUser(
        schemas.UserAccountCreate(
            email="admin@ceonmedia.com",
            password="123",
            profile=schemas.UserProfileCreate(
                username="ceonmedia", first_name="Dayne", last_name="Kolk"
            ),
        ),
        [UserRoleOption.ADMIN],
    ),
    DefaultUser(
        schemas.UserAccountCreate(email="daynekolk@hotmail.com", password="123")
    ),
]


def create_default_user(session: Session, user_to_create: DefaultUser):
    new_user = crud.user.create(session, user_to_create.user_account)
    logger.debug(f"Created new user: {new_user.email}")
    if user_to_create.roles:
        logger.warning(
            f"Assigning roles to user {new_user.email}: {user_to_create.roles}"
        )
        for role in user_to_create.roles:
            crud.access_control.add_role_to_user(
                session, user_account=new_user, role_to_add=role
            )
    session.commit()


def create_default_users(session: Session):
    for default_user in DEFAULT_USERS:
        create_default_user(session, default_user)


def create_admin_user(session: Session):
    # TODO setup admin with env vars
    raise NotImplementedError



import logging

from ceonmedia_db import crud

logger = logging.getLogger(__name__)


def test_get_all_users(default_db_session):
    with default_db_session() as session:
        users_in_db = crud.user.get_all(session)
        logger.warning(
            f"Successfully got {len(users_in_db)} users in db: {users_in_db}"
        )



import logging
import pytest

from sqlmodel import Session

from ceonmedia_db import crud
from ceonmedia_db import models
from ceonmedia_db import schemas
from ceonmedia_db import errors

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel("INFO")

# TODO split this module into three test modules
# test_user_create
# test_user_update
# test_user_get (e.g. get by role (admin, affiliate) etc)


def add_user(default_db_session, user_to_create) -> models.UserAccountTbl:
    logger.info(f"Creating user: {user_to_create.email}")
    user_in_db = crud.user.create(default_db_session, user_to_create)
    logger.debug(f"Successfully added user: {user_in_db}")
    return user_in_db


def delete_user(db: Session, User):
    raise NotImplementedError


# TODO test_add_user_fail_already_exists():
# TODO test_add_user_with_role_success():
# TODO test_add_user_with_role_fail():

# Include deletes in the 'create' module?
# TODO test_delete_user():

# Different module for getting/updating information.
# TODO test_get_user():
# TODO test_get_user_by_email():

NEW_USER_CREATE = schemas.UserAccountCreate(
    email="newly_created_user@hotmail.com", password="123"
)


# Create a simple, regular user
def test_add_user_success(default_db_session):
    with default_db_session() as session:
        # reset_schema(session)
        new_user_in_db = add_user(session, NEW_USER_CREATE)
        got_user = crud.user.get_by_email(session, new_user_in_db.email)
        logger.info(f"Got new user in db: {got_user}")


# Create the same user twice and confirm that a AlreadyExistsInDBError is thrown
def test_add_user_already_exists(default_db_session):
    with default_db_session() as session:
        # reset_schema(session)
        new_user_in_db = add_user(session, NEW_USER_CREATE)
        got_user = crud.user.get_by_email(session, new_user_in_db.email)
        logger.info(f"Got new user in db: {got_user}")
        logger.info("Creating user again...")
        with pytest.raises(errors.AlreadyExistsInDBError):
            add_user(session, NEW_USER_CREATE)
    logger.warning(
        "TODO: Confirm no side-effects on failure? E.g. no child tables created"
    )



from uuid import uuid4, UUID
from datetime import datetime

from sqlmodel import SQLModel, Field, Relationship
from sqlalchemy.dialects.postgresql import JSONB

from sqlalchemy import (
    func,
    text,
)


class ProjectTbl(SQLModel, table=True):
    __tablename__ = "project"

    id: UUID = Field(
        primary_key=True,
        default_factory=uuid4,
        sa_column_kwargs={
            "server_default": text("gen_random_uuid()"),
        },
    )
    owner_id: UUID = Field(foreign_key="user.id", nullable=False)
    title: str = Field(max_length=50)
    description: str
    proj_inputs: dict = Field(sa_type=JSONB, nullable=False)
    created_at: datetime = Field(
        sa_column_kwargs={"server_default": func.now(), "onupdate": func.now()}
    )
    updated_at: datetime = Field(
        sa_column_kwargs={"server_default": func.now(), "onupdate": func.now()}
    )

    owner: "UserTbl" = Relationship(back_populates="user_projects")



import logging
import pytest

from sqlmodel import Session

from ceonmedia_db import crud
from ceonmedia_db import models
from ceonmedia_db import schemas

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel("DEBUG")

# TODO split this module into three test modules
# test_user_create
# test_user_update
# test_user_get (e.g. get by role (admin, affiliate) etc)


def add_project(
    db: Session, project_to_create: schemas.ProjectCreate
) -> models.ProjectTbl:
    logger.info(f"Creating project: {project_to_create.title}")
    project_in_db = crud.project.create(db, project_to_create)
    logger.debug(f"Successfully added project: {project_in_db}")
    return project_in_db


def delete_project(db: Session, User):
    raise NotImplementedError


@pytest.fixture()
def project_to_create(default_db_session) -> schemas.ProjectCreate:
    default_owner_email = "admin@ceonmedia.com"
    with default_db_session() as session:
        default_owner = crud.user.get_by_email(session, default_owner_email)
        logger.debug(f"Got default owner: {default_owner}")
    project_to_create = schemas.ProjectCreate(
        title="Test Project CRUD",
        description="This project was created to test the DB CRUD api",
        proj_inputs=[{"testinput1": "aaa"}],
        owner_uuid=default_owner.uuid,
    )
    return project_to_create


# Create a simple, regular user
def test_create_project_success(default_db_session, project_to_create):
    with default_db_session() as session:
        # reset_schema(session)
        new_project_in_db = add_project(session, project_to_create)
        got_project = crud.project.get(session, new_project_in_db.uuid)
        logger.info(f"Got new project in db: {got_project}")



from uuid import uuid4, UUID

from sqlmodel import SQLModel, Field

from sqlalchemy import (
    text,
)

# Information about goods that can be ordered.


class ProductTbl(SQLModel, table=True):
    __tablename__ = "product"

    id: UUID = Field(
        primary_key=True,
        index=True,
        default_factory=uuid4,
        sa_column_kwargs={"server_default": text("gen_random_uuid()")},
    )
    name: str = Field(max_length=30, nullable=False)
    description: str = Field(max_length=250, nullable=False)
    price_gbp_pence: int = Field(nullable=False)



# Pydantic models, renamed to 'schemas' to avoid being confused with sqlalchemy models.
from typing import Optional

# from pydantic import BaseModel, Field
from uuid import UUID


class ProjectCreate:
    title: str
    proj_inputs: list
    owner_uuid: Optional[UUID]
    description: Optional[str]
    uuid: Optional[UUID] = None

    class Config:
        orm_mode = True


# Remember to add Config orm_mode=True for classes which are reading db entries
# class Config:
#     orm_mode = True



from uuid import UUID

from sqlmodel import SQLModel, Field

# from ceonmedia_db.models.base import Base


class UserStripeTbl(SQLModel, table=True):
    # Map an app user to their stripe customer id.
    __tablename__ = "user_stripe"

    user_account_id: UUID = Field(
        foreign_key="user.id",
        ondelete="CASCADE",
        primary_key=True,
    )
    stripe_customer_id: str = Field(max_length=30, unique=True)



import logging
from uuid import UUID
from sqlalchemy import select
from sqlmodel import Session

from ceonmedia_db.models.submission import SubmissionTbl
from ceonmedia_db import constants

logger = logging.getLogger(constants.LOGGER_NAME)


def create(
    db,
    user_id: str,
    project_id: str,
    submission_id: str | None,
):
    logger.info(f"Received values: {submission_id=}, {user_id=}, {project_id=}")
    new_submission = SubmissionTbl(
        user_account_id=user_id,
        project_id=project_id,
        id=submission_id,
    )
    db.add(new_submission)
    db.commit()
    return new_submission


def get(db: Session, submission_id: UUID):
    # TODO switch to ORM workflow (is select ORM?)
    stmt = select(SubmissionTbl).where(SubmissionTbl.id == submission_id)
    got_submission = db.execute(stmt).scalars().one()
    return got_submission


# def get_by_job_id(db: Session, job_id: UUID) -> SubmissionTbl:
#     stmt = select(SubmissionTbl).where(SubmissionTbl.job_id == job_id)
#     got_submission = db.execute(stmt).scalars().one()
#     return got_submission



import logging

import sqlalchemy
from sqlmodel import Session

from sqlmodel import SQLModel

# Importing the models also means that all of the classes
# which inherit from base are imported (via __init__.py)
# This is necessary for the tables to register their metadata.

# Setup logger
logger = logging.getLogger(__name__)
logger.setLevel("INFO")


def get_live_metadata(session: Session):
    """Get the metadata from the live DB via the engine connection"""
    engine = session.get_bind()
    # Create empty metadata
    live_metadata = sqlalchemy.schema.MetaData()
    # Popoulate it with data from the db
    live_metadata.reflect(bind=engine)
    return live_metadata


def get_python_metadata():
    """
    Returns the metadata as defined by local python modules inheriting from the base class
    Note that classes must have been imported to be 'loaded'
        from ceonmedia_db import models <-- imported classes via __init__.py
    """
    # return models.base.Base.metadata
    return SQLModel.metadata


def clear_data(session: Session):
    logger.warning(f"Clearing all entries in db...")
    live_metadata = get_live_metadata(session)
    for table in reversed(live_metadata.sorted_tables):
        logger.info(f"Removing all entries in table: {table}")
        session.execute(table.delete())
    session.commit()


def print_live_tables(session: Session):
    # Get the metadata from the live DB
    live_metadata = get_live_metadata(session)
    if not live_metadata.tables:
        logger.info(f"\t(None found)")
        return
    for table in live_metadata.tables:
        logger.info(f"\t{table}")


def clear_tables(session):
    """Drop all tables in the DB"""
    logger.warning(f"Dropping all tables in db...")
    live_metadata = get_live_metadata(session)
    engine = session.get_bind()
    live_metadata.drop_all(bind=engine)


def create_tables(session):
    """Load the tables into the live DB as defined by our sqlalchemy models"""
    logger.info("Creating tables...")
    engine = session.get_bind()
    metadata = get_python_metadata()
    metadata.create_all(bind=engine)
    print_live_tables(session)


def reset_schema(session):
    logger.warning(
        "Resetting schema (Dropping all tables and data) and recreating tables..."
    )
    clear_tables(session)
    create_tables(session)



import logging

from sqlalchemy import select
from sqlmodel import Session
from sqlalchemy.exc import SQLAlchemyError

from ceonmedia_db import schemas
from ceonmedia_db import models
from ceonmedia_db import errors
from ceonmedia_db import constants

logger = logging.getLogger(constants.LOGGER_NAME)


# --------------
# User Roles
# --------------
def create_role(db: Session, user_role: schemas.UserRoleOption) -> models.RoleTbl:
    logger.debug(f"Creating new user role: {user_role}")
    new_user_role = models.RoleTbl(name=user_role.value)
    logger.debug(f"Adding new user role to DB: {new_user_role}")
    db.add(new_user_role)

    # -- Commit Changes --
    try:
        db.commit()
        logger.debug(f"Committed new user role to db: {new_user_role.name}")
    except SQLAlchemyError as e:
        msg = f"Failed to create role {user_role}: {e}"
        errors.handle_sql_error(e, msg=msg)

    return new_user_role


def get_all_user_roles(db: Session) -> list[schemas.UserRoleOption]:
    stmt = select(models.RoleTbl).order_by(models.RoleTbl.name)
    result = db.execute(stmt).scalars().all()
    logger.debug(f"result: {result}")
    list_of_enum_instances = [
        schemas.UserRoleOption(user_role.name) for user_role in list(result)
    ]
    return list_of_enum_instances


# --------------
# Permisissions
# --------------
def create_permission(
    db: Session, permission_to_create: schemas.PermissionOption
) -> models.PermissionTbl:
    """Add a new permission to the database"""
    logger.debug(f"Creating new permission: {permission_to_create}")
    new_permission = models.PermissionTbl(name=permission_to_create.value)
    logger.debug(f"Adding new permission to db: {new_permission}")
    db.add(new_permission)

    # -- Commit Changes --
    db.commit()
    logger.debug(f"Committed new permission to db: {new_permission}")

    return new_permission


def get_all_permissions(
    db: Session,
) -> list[schemas.PermissionOption]:
    stmt = select(models.PermissionTbl).order_by(models.PermissionTbl.name)
    result = db.execute(stmt).scalars().all()
    logger.debug(f"result: {result}")
    list_of_enum_instances = [
        schemas.PermissionOption(permission.name) for permission in list(result)
    ]
    return list_of_enum_instances


def get_all_permissions_for_user_role(
    db: Session, user_role: schemas.UserRoleOption
) -> list[schemas.PermissionOption]:
    user_role_in_db = _get_user_role_in_db(db, user_role)
    permissions = user_role_in_db.permissions
    list_of_enums = [
        schemas.PermissionOption(permission.name) for permission in permissions
    ]
    return list_of_enums


# -----------
# Assignments
# -----------
def add_role_to_user(
    db: Session,
    user_account: models.user.UserAccountTbl,
    role_to_add: schemas.UserRoleOption,
):
    user_role_in_db: models.RoleTbl = _get_user_role_in_db(db, role_to_add)
    if user_role_in_db in user_account.user_roles:
        raise errors.AlreadyExistsInDBError(
            f"Tried to add already existing role ({role_to_add}) to user {user_account.email}"
        )
    user_account.user_roles.append(user_role_in_db)
    db.commit()


def add_permission_to_role(
    db: Session,
    user_role: schemas.UserRoleOption,
    permission_to_add: schemas.PermissionOption,
):
    logger.warning(
        f"Adding permission to role {user_role.value}: {permission_to_add.value}"
    )
    user_role_in_db = _get_user_role_in_db(db, user_role)
    permission_in_db = _get_permission_in_db(db, permission_to_add)
    if permission_in_db in user_role_in_db.permissions:
        # The UniqueConstraint on SQLAlchemy association table doesn't seem to raise an
        # error when making duplicate entries (but it does stop duplicates from being created)
        # Therefore, we explicitly check for and raise the error here.
        msg = f"Failed to add permissoin to role {user_role}: {permission_to_add} (already exists)"
        logger.error(msg)
        raise errors.AlreadyExistsInDBError(msg)
    user_role_in_db.permissions.append(permission_in_db)
    db.commit()


# ----
# Helper functions
# ----
# Not intended to be used by the user since the user
# will always pass/receive Enum instances.
def _get_user_role_in_db(
    db: Session, user_role: schemas.UserRoleOption
) -> models.RoleTbl:
    stmt = select(models.RoleTbl).where(models.RoleTbl.name == user_role.value)
    got_user_role = db.execute(stmt).scalars().one()
    logger.debug(f"Returning: {got_user_role}")
    return got_user_role


def _get_permission_in_db(
    db: Session, user_role: schemas.PermissionOption
) -> models.PermissionTbl:
    stmt = select(models.PermissionTbl).where(
        models.PermissionTbl.name == user_role.value
    )
    got_permission = db.execute(stmt).scalars().one()
    logger.debug(f"Returning: {got_permission}")
    return got_permission



from uuid import uuid4
from datetime import datetime, timedelta

from sqlalchemy import (
    String,
    DateTime,
    ForeignKey,
    func,
    text,
)
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.dialects.postgresql import UUID

from ceonmedia_db.models.base import Base


# For temoporary security tokens
def get_valid_until():
    valid_until = datetime.now() + timedelta(hours=1)
    return valid_until


# TODO move this to REDIS DB (more appropriate for ephemeral data, avoids needing cron jobs
# to delete expired data)
class EmailVerificationTokenTbl(Base):
    __tablename__ = "email_verification_token"

    uuid: Mapped[UUID] = mapped_column(
        UUID(as_uuid=True),
        primary_key=True,
        index=True,
        unique=True,
        default=uuid4,
        server_default=text("gen_random_uuid()"),
    )
    user_account_uuid: Mapped[UUID] = mapped_column(
        UUID(as_uuid=True),
        ForeignKey("user_account.uuid", ondelete="CASCADE"),
    )
    email_to_verify: Mapped[str] = mapped_column(String(255))
    created_at: Mapped[DateTime] = mapped_column(
        DateTime(timezone=True), server_default=func.now()
    )
    expires_at: Mapped[DateTime] = mapped_column(
        DateTime(timezone=True), default=get_valid_until
    )
    # user: Mapped["UserAccountTbl"] = relationship()



import logging
from dotenv import load_dotenv

load_dotenv("../secret/.env")

from ceonmedia_db.database import SessionLocal
from ceonmedia_db import db_setup

# Make logging visible
root_logger = logging.getLogger()
root_logger.addHandler(logging.StreamHandler())
root_logger.setLevel("INFO")

logger = logging.getLogger(__name__)


def main():
    db = SessionLocal()
    logger.info(f"Connected to db session: {db}")
    with db as session:
        db_setup.reset.reset_schema(session)

        # Create user_roles
        db_setup.setup_user_roles.push_to_db(session)

        # Read and print
        print()
        db_setup.setup_user_roles.print_user_roles(session)
        db_setup.setup_user_roles.print_permissions(session)
        db_setup.setup_user_roles.print_granted_permissions(session)

        # Create default users
        print()
        logger.info("Creating default users...")
        db_setup.users.create_default_users(session)


if __name__ == "__main__":
    main()



from uuid import uuid4, UUID
from datetime import date, timedelta, datetime

from sqlmodel import SQLModel, Field

import sqlalchemy as sa


def get_valid_until():
    valid_until = date.today() + timedelta(days=30)
    return valid_until


class SubmissionTbl(SQLModel, table=True):
    __tablename__ = "submission"

    id: UUID = Field(
        default_factory=uuid4,
        primary_key=True,
        sa_column_kwargs={"server_default": sa.text("gen_random_uuid()")},
        nullable=False,
        index=True,
    )
    user_account_id: UUID = Field(foreign_key="user.id", ondelete="CASCADE")
    project_id: UUID = Field(foreign_key="project.id", ondelete="CASCADE")
    validated: bool = Field(
        default=False,
        sa_column_kwargs={
            "server_default": "false",
        },
        nullable=False,
    )
    expires_at: datetime = Field(default_factory=get_valid_until)
    created_at: datetime | None = Field(
        default=None, nullable=False, sa_column_kwargs={"server_default": sa.func.now()}
    )
    updated_at: datetime | None = Field(
        default=None,
        sa_type=sa.DateTime(timezone=True),
        sa_column_kwargs={"server_default": sa.func.now(), "onupdate": sa.func.now()},
    )
    # project: "ProjectTbl" = Relationship()



# Test
import logging
from datetime import datetime, timedelta
from datetime import timezone

from sqlmodel import Session
from sqlalchemy import exc
from sqlalchemy import select


from ceonmedia_db import models
from ceonmedia_db import errors

# TODO can proj_crud safely auto-fetch the SessionLocal without it needing to be passed?
# from db.engine import SessionLocal

logger = logging.getLogger(__name__)
# logger.setLevel("DEBUG")


def get(db: Session, session_id: str) -> models.user.UserSessionsTbl:
    logger.info(f"Fetching user for session: {session_id=}")
    stmt = select(models.user.UserSessionsTbl).where(
        models.user.UserSessionsTbl.id == session_id
    )
    try:
        got_session = db.execute(stmt).scalars().one()
    except exc.NoResultFound:
        raise errors.NotFoundInDBError("Could not find session for session_id")
    logger.info(f"got session: {got_session=}")
    expired = got_session.expires_at <= datetime.now(tz=timezone.utc)
    # Get the session
    logger.info(f"{expired=}")
    if expired:
        # TODO create appropriate exception
        logger.warning("TODO: Delete expired sessions")
        raise Exception("Session expired")
    return got_session


# TODO move sessions to redis?
def create(db: Session, user_id: str) -> models.user.UserSessionsTbl:
    # Create new job entry
    user_id = str(user_id)
    time_to_live = timedelta(days=30)
    expires_at = datetime.now(tz=timezone.utc) + time_to_live
    logger.info(f"{expires_at=}")
    new_session = models.user.UserSessionsTbl(user_id=user_id, expires_at=expires_at)
    db.add(new_session)
    db.flush()
    logger.info(f"Created new_session: {new_session.__dict__}")
    return new_session



from uuid import uuid4, UUID
from datetime import datetime

from sqlmodel import SQLModel, Field, Relationship

from sqlalchemy import (
    func,
    text,
)

# from ceonmedia_db.models.base import Base
# from ceonmedia_db.models.user import UserAccountTbl


# Token to allow users to get free preview renders.
# They automatically refresh after a fixed time.
# Spent tokens are automatically refreshed if the user purchases the project.
class ReusablePreviewTokenTbl(SQLModel, table=True):
    __tablename__ = "reusable_preview_token"

    id: UUID = Field(
        default_factory=uuid4,
        primary_key=True,
        index=True,
        sa_column_kwargs={
            "server_default": text("gen_random_uuid()"),
        },
    )
    # Changed to regular id for compatability with oauth workflow, but left uuid naming convention to prevent
    # breakage
    user_account_id: UUID = Field(
        foreign_key="user.id",
        ondelete="CASCADE",
        nullable=False,
    )
    # Remember the project which this token was last spent on.
    project_id: UUID | None = Field(
        foreign_key="project.id",
        ondelete="SET NULL",
    )
    created_at: datetime = Field(sa_column_kwargs={"server_default": func.now()})
    available_at: datetime = Field(sa_column_kwargs={"server_default": func.now()})
    user: "UserTbl" = Relationship(back_populates="reusable_preview_credits")
